// Set 4-byte alignment to have the same placement in both 32- and 64-bit code

__pragma (pack (push, 4))





namespace DsOverAsio {





#define FilterDllName L"directshowoverasio.dll"
#define FilterId "0078e09e-0e91-4686-8783-8cdf7c3f7b56"
#define CaptureFilterClassId "a115c133-13a0-4618-8b75-4990cd522b17"
#define RenderFilterClassId "285bf116-d01e-4a39-a429-0f020bd28766"
#define PropertyPageId "698d4dc1-15ed-46a3-b974-3ec05c865a1b"
#define DsOverAsioCategoryId "3a0d1c06-43cd-4176-a0fc-eb30127b7618"





// Maximum number of supported ASIO channels of each direction
  
static UINT const NumAsioChannelsMax = 32;





//
// Version info
// =============================================================================================
//

struct FilterVersion {

  UINT Major;
  UINT Minor;
  UINT Revision;
  UINT Build;

};





//
// Sample format
// =============================================================================================
//

struct SampleFormatDesc {

  BYTE BitsPerSample;             // Number of meaningful bits per sample
  BYTE BytesPerSample;            // Number of bytes per sample container
  bool FloatingPoint;             // Sample format: IEEE 754 floating point or integer PCM
  bool RightAlign;                // Right aligned (suitable only for integer PCM)

  SampleFormatDesc (UINT const BPS, UINT const BPC = 0, bool const FP = false, bool const RA = false) {
    Fill (BPS, BPC, FP, RA);
  }

  SampleFormatDesc () { Fill (16); }

  void Fill (UINT const BPS, UINT const BPC = 0, bool const FP = false, bool const RA = false) {
    BitsPerSample = static_cast <BYTE> (BPS);
    BytesPerSample = static_cast <BYTE> ((BPC != 0)? BPC : (BPS + 7) / 8);
    FloatingPoint = FP;
    RightAlign = RA;
  }

};





//
// ASIO device/driver entry descriptor
// =============================================================================================
//
// Describes an entry registered in HKLM\Software\ASIO
//
  
struct AsioDeviceEntry {

  wchar_t Name [100];                     // Device/driver name
  GUID ClassId;                           // Driver COM object class ID

};





//
// ASIO buffer parameters
// =============================================================================================
//
  
struct AsioBufferSizes {

  UINT Min;                     // Minimal number of samples per buffer
  UINT Max;                     // Maximal number of samples per buffer
  UINT Preferred;               // Preferred number of samples per buffer
  UINT Granularity;             // Buffer size granularity

};





//
// ASIO channel parameters
// =============================================================================================
//
  
struct AsioChannelParameters {

  UINT NumChannels;             // Number of supported channels (for a direction)
  SampleFormatDesc SampleFormats [NumAsioChannelsMax]; // Channel sample formats (for a direction)

};





//
// ASIO buffering parameters
// =============================================================================================
//
  
struct AsioBufferParameters {

  UINT Latency;                 // Latency in nanoseconds (for a direction). -1 - unknown
  AsioBufferSizes BufferSizes;  // Buffer sizes (for the device)
  UINT ActualBufferSize;        // Actual/current buffer size (for the device)

};





//
// GUID structures to use __uuidof
// =============================================================================================
//
  
struct __declspec (uuid (CaptureFilterClassId)) CaptureFilterClassIdStruct;
struct __declspec (uuid (RenderFilterClassId)) RenderFilterClassIdStruct;
struct __declspec (uuid (PropertyPageId)) PropPageClassIdStruct;
struct __declspec (uuid (DsOverAsioCategoryId)) DsOverAsioCategoryIdStruct;





//
// Filter's control interface
// =============================================================================================
//
  
__interface __declspec (uuid (FilterId)) IDsOverAsio : IUnknown {

  HRESULT GetFilterVersion (FilterVersion & Ver);
  HRESULT EnumerateAsioDevices (AsioDeviceEntry DevArray [], UINT const NumEntries, UINT & NumDevices);
  HRESULT SetAsioDriverWindow (HWND const Win);
  HRESULT SelectAsioDevice (GUID const & ClassId);
  HRESULT ShowAsioControlPanel ();
  HRESULT GetChannelParameters (AsioChannelParameters & Params);
  HRESULT GetBufferParameters (AsioBufferParameters & Params);
  HRESULT QuerySamplingRateSupport (UINT const Rate);
  HRESULT GetActualSamplingRate (UINT & Rate);
  HRESULT SetSamplingRate (UINT const Rate);
  HRESULT SetDeviceBufferSize (UINT const Samples);
  HRESULT SetPinBufferSize (UINT const Samples, UINT const NumParts = 0);
  HRESULT SetPinSampleType (SampleFormatDesc const Type);

};





} // namespace





__pragma (pack (pop))
