#include "stdafx.h"





// #############################################################################################
//
// Macros
// =============================================================================================
//
// #############################################################################################





// Debugging parameters

#define ConnectPins 1
#define UseMediaControl 1
#define PauseBeforeRun 0
#define RunGraph 1





// #############################################################################################
//
// Classes
// =============================================================================================
//
// #############################################################################################





struct DsAudioFilter {

  wchar_t Name [50];
  wchar_t FullName [MAX_PATH];
  GUID ClassId;

};


struct DsAudioFilterList {

  UINT NumFilters;
  DsAudioFilter Filters [50];

};





// #############################################################################################
//
// Types
// =============================================================================================
//
// #############################################################################################





typedef HRESULT WINAPI DllSetDebugPrintLevel_t (UINT const Level);





// #############################################################################################
//
// Constants
// =============================================================================================
//
// #############################################################################################





static PCTSTR const StateNames [] = { L"stop", L"pause", L"run" };



static UINT const SamplingRateList [] = {

  5000, 8000, 11025, 22050, 44100, 48000, 88200, 96000, 192000, UINT_MAX

};

static PCTSTR const SampleTypeList [] = {

  L"PCM", L"Float", nullptr

};

static PCTSTR const SampleAlignmentList [] = {

  L"left", L"right", nullptr

};


static UINT const BitsPerSampleList [] = {

  8, 16, 18, 20, 22, 24, 32, UINT_MAX

};

static UINT const BytesPerContainerList [] = {

  0, 1, 2, 3, 4, 8, UINT_MAX

};

static UINT const SamplesPerBufferList [] = {

  128, 192, 256, 384, 512, 768, 1024, 1536, 2048, UINT_MAX

};

static UINT const NumBufferPartsList [] = {

  1, 2, 4, 8, 16, UINT_MAX

};



static UINT const AsioChannelControls [] = {

  IDC_AsioChannels, IDC_AsioSampleType, IDC_AsioBps, IDC_AsioBpc, IDC_AsioSampleAlignment,

  0

};

static UINT const AsioBufferControls [] = {

  IDC_AsioBufferMin, IDC_AsioBufferMax, IDC_AsioBufferPref, IDC_AsioBufferGran,
  IDC_AsioBufferActual, IDC_AsioLatency,

  0

};

static UINT const AsioInputControls [] = {

  IDC_AsioDevice, IDC_AsioSamplesPerBuffer, IDC_AsioControlPanel,

  0

};

static UINT const NonRealtimeControls [] = {

  IDC_CaptureFilter,
  IDC_SamplingRate, IDC_PinSampleType, IDC_PinBitsPerSample,
  IDC_PinSamplesPerBuffer, IDC_PinBufferParts,
  IDC_RenderFilter,

  0

};

static UINT const AsioNonRealtimeControls [] = {

  IDC_AsioDevice, IDC_AsioSamplesPerBuffer,

  0

};





// #############################################################################################
//
// Variables
// =============================================================================================
//
// #############################################################################################





#if _Debug

  UINT DebugPrintLevel = 5;

#endif





HINSTANCE AppInst = nullptr;
HWND MainDlg = 0;

DsAudioFilterList CaptureFilters;
DsAudioFilterList RenderFilters;

HMODULE FilterDll = nullptr;
UINT FilterDllRefs = 0;
DllSetDebugPrintLevel_t * DllSetDebugPrintLevel_a = nullptr;
LPFNGETCLASSOBJECT DllGetClassObject_a = nullptr;

IBaseFilter * RenderFilter = nullptr;
bool RenderFilterAdded = false;
IPin * InputPin = nullptr;
IDsOverAsio * RenderFilterCI = nullptr;

IBaseFilter * CaptureFilter = nullptr;
bool CaptureFilterAdded = false;
IPin * OutputPin = nullptr;
IDsOverAsio * CaptureFilterCI = nullptr;

bool PinsConnected = false;

AsioDeviceEntry AsioDevices [AsioDevicesMax];
UINT NumAsioDevices = 0;

IGraphBuilder * GraphBuilder = nullptr;
IMediaControl * MediaControl = nullptr;



bool Started = false;





// #############################################################################################
//
// Functions
// =============================================================================================
//
// #############################################################################################





//
// Enumerate PCM capture/render filters
// =============================================================================================
//

HRESULT EnumeratePcmFilters (bool const Capture) {

  HRESULT FinalRes;

  PCTSTR const Type = Capture? L"capture" : L"render";

  DsAudioFilterList & FilterList = Capture? CaptureFilters : RenderFilters;

  FilterList.NumFilters = 0;

  IFilterMapper2 * FilterMapper = nullptr;
  ICreateDevEnum * DevEnum = nullptr;

  IEnumMoniker * EnumMoniker = nullptr;

  do {

    HRESULT Res = E_I_PROCESSING_NOT_FINISHED;

    Debug (6, L"Enumerating PCM %s filters", Type);


#if 1

    GUID const & Category = Capture? (
      CLSID_AudioInputDeviceCategory
      //AM_KSCATEGORY_CAPTURE
    ) : (
      CLSID_AudioRendererCategory
      //AM_KSCATEGORY_RENDER
    );

    Res = CreateObjectInstance (CLSID_SystemDeviceEnum, __uuidof (ICreateDevEnum), DevEnum);

    if (Res != S_OK) {

      Error (L"Cannot create device enumerator (%s)", ErrorText (Res));

      FinalRes = Res;

      break;

    }

    Debug (7, L"ICreateDevEnum: %p", DevEnum);

    Assert (DevEnum != nullptr);

    Res = DevEnum->CreateClassEnumerator (
      Category, &EnumMoniker, 
      (0
        //| CDEF_DEVMON_FILTER
        //| CDEF_DEVMON_PNP_DEVICE
      )
    );

    if (Res != S_OK) {

      Error (L"Cannot enumerate PCM %s filters (%s)", Type, ErrorText (Res));

      FinalRes = Res;

      break;

    }

#else

    Res = CreateObjectInstance (CLSID_FilterMapper2, __uuidof (IFilterMapper2), FilterMapper);

    if (Res != S_OK) {

      Error (L"Cannot create filter mapper (%s)", ErrorText (Res));

      FinalRes = Res;

      break;

    }

    Debug (7, L"IFilterMapper2: %p", FilterMapper);

    Assert (FilterMapper != nullptr);

    DisableWarningsPush (Warnings_LocalStaticNotThreadSafe)

    static GUID const Types [] = { MEDIATYPE_Audio, MEDIASUBTYPE_PCM };

    PopWarnings ()

    Res = FilterMapper->EnumMatchingFilters (
      &EnumMoniker,
      0,
      false,
      MERIT_DO_NOT_USE,
      !Capture,
      Capture? 0 : NumArrayElems (Types) / 2,
      Capture? nullptr : Types,
      nullptr,
      nullptr,
      !Capture,
      Capture,
      Capture? NumArrayElems (Types) / 2 : 0,
      Capture? Types : nullptr,
      nullptr,
      nullptr
    );

    if (Res != S_OK) {

      Error (L"Cannot enumerate filters (%s)", ErrorText (Res));

      FinalRes = Res;

      break;

    }

#endif

    Assert (EnumMoniker != nullptr);



    IMoniker * Moniker = nullptr;
    ULONG Fetched = 0;

    while ((Res = EnumMoniker->Next (1, &Moniker, /*nullptr*/&Fetched)) == S_OK) {

      Assert (Moniker != nullptr);

      IPropertyBag * Bag = nullptr;
      IBindCtx * BindCtx = nullptr;
      LPOLESTR DisplayNameVar = nullptr;

      do {

		    Res = CreateBindCtx (0, &BindCtx);

        if (Res != S_OK) {

          Debug (9, L"Cannot create bind context (%s)", ErrorText (Res));

          break;

        }

        PCTSTR FullName = L"";

        Res = Moniker->GetDisplayName (BindCtx, nullptr, &DisplayNameVar);

        if (Res == S_OK) {

          FullName = DisplayNameVar;

        } else {

          Debug (9, L"Cannot get display name (%s)", ErrorText (Res));

          break;

        }

        Res = Moniker->BindToStorage (
          nullptr, nullptr, __uuidof (IPropertyBag), PointerCast (void **, &Bag)
        );

        if (Res != S_OK) {

          Error (L"Cannot bind moniker to the storage (%s)", ErrorText (Res));

          break;

        }

        Assert (Bag != nullptr);

        VARIANT NameVar;
        VariantInit (&NameVar);

        Res = Bag->Read (L"FriendlyName", &NameVar, nullptr);

        if (Res != S_OK) {

          Error (L"Cannot get filter friendly name (%s)", ErrorText (Res));

          break;

        }

        Assert (NameVar.vt == VT_BSTR);

        VARIANT ClsIdVar;
        VariantInit (&ClsIdVar);

        Res = Bag->Read (L"CLSID", &ClsIdVar, nullptr);

        if (Res != S_OK) {

          Error (L"Cannot get filter CLSID (%s)", ErrorText (Res));

          break;

        }

        Assert (ClsIdVar.vt == VT_BSTR);

        Debug (
          7,
          L"%s filter %u: %s (\"%s\"), %s",
          Type, FilterList.NumFilters + 1, NameVar.bstrVal, FullName, ClsIdVar.bstrVal
        );

        GUID ClassId;

        Res = CLSIDFromString (ClsIdVar.bstrVal, &ClassId);

        if (Res != S_OK) {

          Debug (
            9,
            L"Cannot convert class ID \"%s\" to binary form (%s)",
            ClsIdVar.bstrVal, ErrorText (Res)
          );

          break;

        }

        if (FilterList.NumFilters == NumArrayElems (FilterList.Filters)) {

          Error (
            L"Number of %s registered filters is greater than %u",
            NumArrayElems (FilterList.Filters)
          );

          Res = HRESULT_FROM_WIN32 (ERROR_TOO_MANY_NAMES);

          break;

        }

        DsAudioFilter & FilterEntry = FilterList.Filters [FilterList.NumFilters];

        StringCchCopy (FilterEntry.Name, NumArrayElems (FilterEntry.Name), NameVar.bstrVal);
        StringCchCopy (FilterEntry.FullName, NumArrayElems (FilterEntry.FullName), FullName);
        FilterEntry.ClassId = ClassId;

        ++FilterList.NumFilters;

        VariantClear (&ClsIdVar);
        VariantClear (&NameVar);

      } OnlyOnce;

      ReleaseObject (Bag IfDebug (, L"IPropertyBag %p", Bag));
      ReleaseObject (BindCtx IfDebug (, L"IBindCtx %p", BindCtx));

      if (DisplayNameVar != nullptr) {

        Debug (6, L"Freeing OLESTR %p", DisplayNameVar);
        CoTaskMemFree (DisplayNameVar);
        DisplayNameVar = nullptr;

      }

      ReleaseObject (Moniker IfDebug (, L"IMoniker %p", Moniker));

    } // EnumMoniker loop




    if (FAILED (Res)) {

      FinalRes = Res;

      break;

    }

    if (FilterList.NumFilters == 0) {

      Error (L"No PCM %s filters found", Type);

      FinalRes = HRESULT_FROM_WIN32 (ERROR_EMPTY);

      break;

    }

    FinalRes = S_OK;

    Debug (7, L"Found %u PCM %s filters", FilterList.NumFilters, Type);

  } OnlyOnce;

  ReleaseObject (EnumMoniker IfDebug (, L"IEnumMoniker %p", EnumMoniker));
  ReleaseObject (DevEnum IfDebug (, L"IDevEnum %p", DevEnum));
  ReleaseObject (FilterMapper IfDebug (, L"IFilterMapper %p", FilterMapper));

  return FinalRes;

}





//
// Instantiate a filter by its full name (moniker's display name)
// =============================================================================================
//

HRESULT InstantiateFilterByFullName (PCTSTR const FullName, IBaseFilter * & Filter) {

  HRESULT FinalRes;

	IBindCtx * BindCtx = nullptr;
  IMoniker * Moniker = nullptr;

  do {

    HRESULT Res;

#if 1

    BIND_OPTS BO;

    BO.cbStruct = sizeof (BO);
    BO.grfFlags = 0;
    BO.grfMode = STGM_READ | STGM_SHARE_DENY_NONE;
    BO.dwTickCountDeadline = 5000;

    Res = CoGetObject (FullName, &BO, __uuidof (IBaseFilter), PointerCast (void **, &Filter));

    if (Res != S_OK) {

      Error (L"Cannot create filter by full name \"%s\" (%s)", FullName, ErrorText (Res));

      FinalRes = Res;

      break;

    }

#else
    
    Res = CreateBindCtx (0, &BindCtx);

    if (Res != S_OK) {

      Error (L"Cannot create bind context for \"%s\" (%s)", FullName, ErrorText (Res));

      FinalRes = Res;

      break;

    }

		ULONG Eaten = 0;

    Res = MkParseDisplayName (BindCtx, FullName, &Eaten, &Moniker);

    if (Res != S_OK) {

      Error (L"Cannot parse filter's full name \"%s\" (%s)", FullName, ErrorText (Res));

      FinalRes = Res;

      break;

    }

	  Res = Moniker->BindToObject (
      BindCtx, nullptr, IID_IBaseFilter, PointerCast (void **, &Filter)
    );

    if (Res != S_OK) {

      Error (L"Cannot bind moniker to filter \"%s\" (%s)", FullName, ErrorText (Res));

      FinalRes = Res;

      break;

    }

#endif

    FinalRes = S_OK;

  } OnlyOnce;

  ReleaseObject (Moniker IfDebug (, L"IMoniker %p", Moniker));
  ReleaseObject (BindCtx IfDebug (, L"IBindCtx %p", BindCtx));

  return FinalRes;

}





//
// Find a PCM pin
// =============================================================================================
//

HRESULT FindPcmPin (IBaseFilter * const Filter, bool const Capture, IPin * & ResultingPin) {

  PCTSTR const Type = Capture? L"capture" : L"render";

  Debug (7, L"Finding %s PCM pin", Type);

  Assert (Filter != nullptr);

  HRESULT FinalRes;

  ResultingPin = nullptr;

  IEnumPins * Pins = nullptr;
  IPin * FoundPin = nullptr;

  do {

    HRESULT Res = E_FAIL;

    Res = Filter->EnumPins (&Pins);

    if (Res != S_OK) {

      Error (L"Cannot enumerate %s filter pins (%s)", Type, ErrorText (Res));

      FinalRes = Res;

      break;

    }

    IPin * Pin = nullptr;

    for (UINT i = 0; (Res = Pins->Next (1, &Pin, nullptr)) == S_OK; ++i) {

      Unreferenced (Capture);

      PWSTR PinId = nullptr;

      do {

        Res = Pin->QueryId (&PinId);

        if (Res != S_OK) {

          Error (L"Cannot get %s filter pin %u ID (%s)", Type, i + 1, ErrorText (Res));

          break;

        }

        PIN_INFO Info;

        Res = Pin->QueryPinInfo (&Info);

        if (Res != S_OK) {

          Error (L"Cannot get %s filter pin %u info (%s)", Type, i + 1, ErrorText (Res));

          break;

        }

        Assert (Info.pFilter == Filter);

        ReleaseObject (Info.pFilter IfDebug (, L"Info.pFilter %p", Info.pFilter));

        Debug (
          7,
          L"%s filter pin %u: type=%s, ID=%s, name=%s",
          Type, i + 1, (Info.dir == PINDIR_INPUT)? L"input" : L"output",
          PinId, Info.achName
        );

        if ((Info.dir == PINDIR_OUTPUT) == Capture) {

          Debug (7, L"Found matching %s filter pin %u \"%s\"", Type, i + 1, Info.achName);

          Assert (Res == S_OK);

          if (FoundPin == nullptr) {

            FoundPin = Pin;

          } else {

            Debug (7, L"Already have a matching pin of %s filter", Type);

          }

          break;

        }

      } OnlyOnce;

      if (PinId != nullptr) {

        CoTaskMemFree (PinId); // Cannot use our delete with foreign allocations
        PinId = nullptr;

      }

      if (FoundPin != nullptr) {

        Assert (Res == S_OK);

        // Exiting the loop keeping pin reference

        break;

      }

      ReleaseObject (Pin IfDebug (, L"IPin %p", Pin));

      if (Res != S_OK) {

        break;

      }

    } // pin loop

    if (Res != S_OK) {

      FinalRes = Res;

      break;

    }

    if (FoundPin == nullptr) {

      FinalRes = HRESULT_FROM_WIN32 (ERROR_NO_MATCH);

      Error (L"PCM pin of %s filter not found", Type);

      break;

    }

    ResultingPin = FoundPin;

    FinalRes = S_OK;

  } OnlyOnce;

  ReleaseObject (Pins IfDebug (, L"IEnumPins %p", Pins));

  return FinalRes;

}





//
// Fill a filter list
// =============================================================================================
//
// If template agrument is nonzero, tries to match filter name. If there is a match, sets
// current position to it. Otherwise, sets current position to zero (our ASIO capture filter).
//

void FillFilterList (bool const Capture, PCTSTR const NameTpl) {

  UINT const CtrlId = static_cast <UINT> (Capture? IDC_CaptureFilter : IDC_RenderFilter);

  DlgMsg (CtrlId, CB_RESETCONTENT, 0, 0);

  DlgMsg (CtrlId, CB_ADDSTRING, 0, reinterpret_cast <LPARAM> (L"ASIO filter (direct DLL loading)"));

  DsAudioFilterList const & Filters = Capture? CaptureFilters : RenderFilters;

  UINT Idx = UINT (-1);

  for (UINT i = 0; i < Filters.NumFilters; ++i) {

    DlgMsg (CtrlId, CB_ADDSTRING, 0, reinterpret_cast <LPARAM> (Filters.Filters [i].Name));

    if (
      Idx == UINT (-1)
      && NameTpl != nullptr && IsTemplateMatching (NameTpl, Filters.Filters [i].Name, true)
    ) {

      Idx = i;

    }

  }

  if (Idx != UINT (-1)) {

    Debug (6, L"%s found at %u", (NameTpl != nullptr)? NameTpl : L"-notemplate-", Idx);

    DlgMsg (CtrlId, CB_SETCURSEL, Idx + 1, 0);

  } else {

    Debug (6, L"%s not found", (NameTpl != nullptr)? NameTpl : L"-notemplate-");

    DlgMsg (CtrlId, CB_SETCURSEL, 0, 0);

  }

}





//
// Enable/disable non-realtime controls
// =============================================================================================
//

void EnableNonRealtimeControls (bool const Enable) {

  EnableControls (NonRealtimeControls, Enable);

  if (CaptureFilterCI != nullptr || RenderFilterCI != nullptr) {

    EnableControls (AsioNonRealtimeControls, Enable);

  }

}





//
// Enable/disable ASIO controls
// =============================================================================================
//

void EnableAsioControls (bool const Enable) {

  EnableControls (AsioInputControls, Enable);

}





//
// Open PCM filter
// =============================================================================================
//

HRESULT OpenPcmFilter (bool const Capture, PCTSTR const FilterNameTpl, IBaseFilter * & Filter) {

  Unreferenced (Capture && FilterNameTpl && Filter);

  return E_NOTIMPL;

}





//
// Unload ASIO filter DLL
// =============================================================================================
//

void UnloadAsioFilterDll () {

  Assert (FilterDllRefs == 0);
  Assert (!Started && !PinsConnected);
  Assert (RenderFilterCI == nullptr && CaptureFilterCI == nullptr);
  Assert (GraphBuilder == nullptr && MediaControl == nullptr);

  if (FilterDll != nullptr) {

    Debug (7, L"Unloading filter DLL");

    DllGetClassObject_a = nullptr;
    DllSetDebugPrintLevel_a = nullptr;

    if (FreeLibrary (FilterDll)) {

      Debug (8, L"Filter DLL unloaded");

    } else {

      HRESULT const Res = HRESULT_FROM_WIN32 (GetLastError ());

      Error (L"Cannot unload capture filter DLL (%s)", ErrorText (Res));

    }

    FilterDll = nullptr;

  }

}





//
// Load ASIO filter DLL
// =============================================================================================
//

HRESULT LoadAsioFilterDll () {

  HRESULT FinalRes;

  do {

    HRESULT Res = E_FAIL;

    Assert (FilterDllRefs == 1);

    PCTSTR const DllPath = (
      L"..\\..\\..\\filter\\"
#if _Debug
      L"debug"
#else
      L"release"
#endif
      L"\\"
#if _64BitEnv
      L"x64"
#else
      L"win32"
#endif
      L"\\" FilterDllName
    );

    Assert (FilterDll == nullptr);

    Debug (7, L"Loading filter DLL \"%s\"", DllPath);

    FilterDll = LoadLibrary (DllPath);

    if (FilterDll == nullptr) {

      Res = HRESULT_FROM_WIN32 (GetLastError ());
      
      Error (L"Cannot load filter DLL \"%s\" (%s)", DllPath, ErrorText (Res));

      FinalRes = Res;

      break;

    }

    Debug (8, L"Filter DLL loaded");

    Assert (DllSetDebugPrintLevel_a == nullptr);

    DllSetDebugPrintLevel_a = (
      PointerCast (DllSetDebugPrintLevel_t *, GetProcAddress (FilterDll, "DllSetDebugPrintLevel"))
    );

    if (DllSetDebugPrintLevel_a == nullptr) {

      Res = HRESULT_FROM_WIN32 (GetLastError ());
      
      Error (L"Cannot get address of DllSetDebugPrintLevel (%s)", ErrorText (Res));

      FinalRes = Res;

      break;

    }

    DllSetDebugPrintLevel_a (DllDebugPrintLevel);

    Assert (DllGetClassObject_a == nullptr);

    DllGetClassObject_a = (
      PointerCast (LPFNGETCLASSOBJECT, GetProcAddress (FilterDll, "DllGetClassObject"))
    );

    if (DllGetClassObject_a == nullptr) {

      Res = HRESULT_FROM_WIN32 (GetLastError ());
      
      Error (L"Cannot get address of DllGetClassObject (%s)", ErrorText (Res));

      FinalRes = Res;

      break;

    }

    FinalRes = S_OK;

  } OnlyOnce;

#if 0

  if (FinalRes != S_OK) {

    UnloadAsioFilterDll ();

  }

#endif

  return FinalRes;

}





void CloseAsio () {

  Assert (RenderFilterCI == nullptr && CaptureFilterCI == nullptr);

  EnableAsioControls (false);

  UnloadAsioFilterDll ();

  NumAsioDevices = 0;

  DlgMsg (IDC_AsioDevice, CB_RESETCONTENT);
  ClearControls (AsioChannelControls);
  ClearControls (AsioBufferControls);

  DlgMsg (IDC_AsioDevice, CB_RESETCONTENT);

  Debug (7, L"ASIO closed");

}





void CloseAsioDir () {

  Debug (7, L"Closing ASIO direction, RC=%u", FilterDllRefs);

  Assert (FilterDllRefs >= 1 && FilterDllRefs <= 2);

  if (--FilterDllRefs == 0) {

    Assert (RenderFilterCI == nullptr && CaptureFilterCI == nullptr);

    CloseAsio ();

  }

  Assert (FilterDllRefs <= 1);

  Debug (8, L"ASIO direction closed, RC=%u", FilterDllRefs);

}





HRESULT OpenAsioDir () {

  HRESULT FinalRes;

  do {

    Debug (7, L"Opening ASIO direction, RC=%u", FilterDllRefs);

    Assert (FilterDllRefs <= 1);

    HRESULT Res;

    if (++FilterDllRefs == 1) {

      Res = LoadAsioFilterDll ();

      if (Res != S_OK) {

        FinalRes = Res;

        break;

      }

    }

    Assert (FilterDll != nullptr);
    Assert (DllSetDebugPrintLevel_a != nullptr);
    Assert (DllGetClassObject_a != nullptr);

    Debug (7, L"ASIO direction open, RC=%u", FilterDllRefs);

    Assert (FilterDllRefs >= 1 && FilterDllRefs <= 2);

    FinalRes = S_OK;

  } OnlyOnce;

  if (FinalRes != S_OK) {

    CloseAsioDir ();

  }

  return FinalRes;

}





//
// Instantiate ASIO capture filter
// =============================================================================================
//

HRESULT InstantiateAsioCaptureFilter () {

  Assert (CaptureFilter == nullptr && OutputPin == nullptr && CaptureFilterCI == nullptr);

  HRESULT FinalRes;

  bool AsioDirOpen = false;

  IClassFactory * Factory = nullptr;

  do {

    HRESULT Res;

    Res = OpenAsioDir ();

    if (Res != S_OK) {

      FinalRes = Res;

      break;

    }

    AsioDirOpen = true;

    Debug (7, L"Requsting capture filter class factory from the DLL");

    Assert (DllGetClassObject_a != nullptr);

    Res = DllGetClassObject_a (
      __uuidof (CaptureFilterClassIdStruct), __uuidof (IClassFactory),
      PointerCast (void **, &Factory)
    );

    if (Res != S_OK) {

      Error (L"Cannot create ASIO capture filter class factory object (%s)", ErrorText (Res));

      FinalRes = Res;

      break;

    }

    Assert (Factory != nullptr && GetObjectRefCount (Factory) == 1);

    IBaseFilter * AsioCaptFilter = nullptr;

    Res = Factory->CreateInstance (
      nullptr,
      __uuidof (IBaseFilter),
      PointerCast (void **, &AsioCaptFilter)
    );

    if (Res != S_OK) {

      Error (L"Cannot create ASIO capture filter object (%s)", ErrorText (Res));

      FinalRes = Res;

      break;

    }

    Assert (GetObjectRefCount (AsioCaptFilter) == 1);

    CaptureFilter = AsioCaptFilter;

    FinalRes = S_OK;

  } OnlyOnce;

  ReleaseObject (Factory IfDebug (, L"capture filter factory %p", Factory));

  if (FinalRes != S_OK && AsioDirOpen) {

    CloseAsioDir ();

  }

  return FinalRes;

}





//
// Get ASIO capture filter control interface
// =============================================================================================
//

HRESULT GetCaptureFilterControlInterface () {

  HRESULT FinalRes;

  do {

    HRESULT Res;

    Assert (CaptureFilter != nullptr && CaptureFilterCI == nullptr);

    Res = CaptureFilter->QueryInterface (
      __uuidof (IDsOverAsio), PointerCast (void **, &CaptureFilterCI)
    );

    if (Res != S_OK) {
      
      if (Res == E_NOINTERFACE) {

        Debug (9, L"Capture filter has no ASIO control interface");

      } else {

        Error (L"Cannot get ASIO capture filter control interface (%s)", ErrorText (Res));

      }

      FinalRes = Res;

      break;

    }

    Assert (CaptureFilterCI != nullptr);

    Debug (8, L"ASIO capture filter control interface: %p", CaptureFilterCI);

    FinalRes = S_OK;

  } OnlyOnce;

  return FinalRes;

}






//
// Instantiate currently selected capture filter
// =============================================================================================
//

HRESULT InstantiateCurrentCaptureFilter () {

  Assert (CaptureFilter == nullptr && OutputPin == nullptr && CaptureFilterCI == nullptr);

  HRESULT FinalRes;

  do {

    HRESULT Res;

    LRESULT LRes;

    LRes = DlgMsg (IDC_CaptureFilter, CB_GETCURSEL);

    Assert (LRes != CB_ERR);

    UINT const CaptureSourceIdx = static_cast <UINT> (LRes);

    if (CaptureSourceIdx == 0) {

      // Instantiate our ASIO capture filter

      Res = InstantiateAsioCaptureFilter ();

      if (Res != S_OK) {

        FinalRes = Res;

        break;

      }

    } else {

      // Instantiate a standard DS capture filter

      UINT const CaptureFilterIdx = CaptureSourceIdx - 1;

      Assert (CaptureFilterIdx < CaptureFilters.NumFilters);

      DsAudioFilter const & CaptureFilterEntry = CaptureFilters.Filters [CaptureFilterIdx];

      Debug (
        7,
        L"Instantiating capture filter %s (%s)",
        CaptureFilterEntry.Name, CaptureFilterEntry.FullName
      );

      Res = InstantiateFilterByFullName (CaptureFilterEntry.FullName, CaptureFilter);

      if (Res != S_OK) {

        FinalRes = Res;

        break;

      }

    }


    Debug (7, L"Capture filter: %p", CaptureFilter);

    Assert (CaptureFilter != nullptr);



    Res = GetCaptureFilterControlInterface ();

    if (Res != S_OK && Res != E_NOINTERFACE) {

      FinalRes = Res;

      break;

    }



    FinalRes = S_OK;

  } OnlyOnce;

  return FinalRes;

}





//
// Instantiate ASIO render filter
// =============================================================================================
//

HRESULT InstantiateAsioRenderFilter () {

  Assert (RenderFilter == nullptr && InputPin == nullptr && RenderFilterCI == nullptr);

  HRESULT FinalRes;

  bool AsioDirOpen = false;

  IClassFactory * Factory = nullptr;

  do {

    HRESULT Res;

    Res = OpenAsioDir ();

    if (Res != S_OK) {

      FinalRes = Res;

      break;

    }

    AsioDirOpen = true;

    Debug (7, L"Requsting render filter class factory from the DLL");

    Assert (DllGetClassObject_a != nullptr);

    Res = DllGetClassObject_a (
      __uuidof (RenderFilterClassIdStruct), __uuidof (IClassFactory),
      PointerCast (void **, &Factory)
    );

    if (Res != S_OK) {

      Error (L"Cannot create ASIO render filter class factory object (%s)", ErrorText (Res));

      FinalRes = Res;

      break;

    }

    Assert (Factory != nullptr && GetObjectRefCount (Factory) == 1);

    IBaseFilter * AsioRendFilter = nullptr;

    Res = Factory->CreateInstance (
      nullptr,
      __uuidof (IBaseFilter),
      PointerCast (void **, &AsioRendFilter)
    );

    if (Res != S_OK) {

      Error (L"Cannot create ASIO render filter object (%s)", ErrorText (Res));

      FinalRes = Res;

      break;

    }

    Assert (GetObjectRefCount (AsioRendFilter) == 1);

    RenderFilter = AsioRendFilter;

    FinalRes = S_OK;

  } OnlyOnce;

  ReleaseObject (Factory IfDebug (, L"render filter factory %p", Factory));

  if (FinalRes != S_OK && AsioDirOpen) {

    CloseAsioDir ();

  }

  return FinalRes;

}





//
// Get ASIO render filter control interface
// =============================================================================================
//

HRESULT GetRenderFilterControlInterface () {

  HRESULT FinalRes;

  do {

    HRESULT Res;

    Assert (RenderFilter != nullptr && RenderFilterCI == nullptr);

    Res = RenderFilter->QueryInterface (
      __uuidof (IDsOverAsio), PointerCast (void **, &RenderFilterCI)
    );

    if (Res != S_OK) {
      
      if (Res == E_NOINTERFACE) {

        Debug (9, L"Render filter has no ASIO control interface");

      } else {

        Error (L"Cannot get ASIO render filter control interface (%s)", ErrorText (Res));

      }

      FinalRes = Res;

      break;

    }

    Assert (RenderFilterCI != nullptr);

    Debug (8, L"ASIO render filter control interface: %p", RenderFilterCI);

    FinalRes = S_OK;

  } OnlyOnce;

  return FinalRes;

}






//
// Instantiate currently selected render filter
// =============================================================================================
//

HRESULT InstantiateCurrentRenderFilter () {

  Assert (RenderFilter == nullptr && InputPin == nullptr && RenderFilterCI == nullptr);

  HRESULT FinalRes;

  do {

    HRESULT Res;

    LRESULT LRes;

    LRes = DlgMsg (IDC_RenderFilter, CB_GETCURSEL);

    Assert (LRes != CB_ERR);

    UINT const RenderSourceIdx = static_cast <UINT> (LRes);

    if (RenderSourceIdx == 0) {

      // Instantiate our ASIO render filter

      Res = InstantiateAsioRenderFilter ();

      if (Res != S_OK) {

        FinalRes = Res;

        break;

      }

    } else {

      // Instantiate a standard DS render filter

      UINT const RenderFilterIdx = RenderSourceIdx - 1;

      Assert (RenderFilterIdx < RenderFilters.NumFilters);

      DsAudioFilter const & RenderFilterEntry = RenderFilters.Filters [RenderFilterIdx];

      Debug (
        7,
        L"Instantiating render filter %s (%s)",
        RenderFilterEntry.Name, RenderFilterEntry.FullName
      );

      Res = InstantiateFilterByFullName (RenderFilterEntry.FullName, RenderFilter);

      if (Res != S_OK) {

        FinalRes = Res;

        break;

      }

    }


    Debug (7, L"Render filter: %p", RenderFilter);

    Assert (RenderFilter != nullptr);



    Res = GetRenderFilterControlInterface ();

    if (Res != S_OK && Res != E_NOINTERFACE) {

      FinalRes = Res;

      break;

    }



    FinalRes = S_OK;

  } OnlyOnce;

  return FinalRes;

}





//
// Fill the list of ASIO devices
// =============================================================================================
//

HRESULT FillAsioDevicesList () {

  IDsOverAsio * FilterCI = (CaptureFilterCI != nullptr)? CaptureFilterCI : RenderFilterCI;

  Assert (FilterCI != nullptr);

  HRESULT FinalRes;

  do {

    HRESULT Res;

    Res = FilterCI->EnumerateAsioDevices (AsioDevices, AsioDevicesMax, NumAsioDevices);

    if (Res != S_OK) {

      Error (L"Cannot get ASIO device list (%s)", ErrorText (Res));

      FinalRes = Res;

      break;

    }

    Debug (7, L"Filling %u ASIO devices", NumAsioDevices);

    Assert (DlgMsg (IDC_AsioDevice, CB_GETCOUNT) == 0);

    for (UINT i = 0; i < NumAsioDevices; ++i) {

      Debug (6, L"ASIO device %u: \"%s\"", i + 1, AsioDevices [i].Name);

      DlgMsg (IDC_AsioDevice, CB_ADDSTRING, 0, reinterpret_cast <LPARAM> (AsioDevices [i].Name));

    }

    DlgMsg (IDC_AsioDevice, CB_SETCURSEL, 0);

    EnableAsioControls (true);

    FinalRes = S_OK;

  } OnlyOnce;

  return FinalRes;

}





//
// Query ASIO filter and fill channel dialog controls
// =============================================================================================
//

HRESULT PopulateAsioChannelInfo () {

  HRESULT FinalRes;

  do {

    IDsOverAsio * FilterCI = (CaptureFilterCI != nullptr)? CaptureFilterCI : RenderFilterCI;

    if (FilterCI == nullptr) {

      Debug (6, L"No filter control interface to populate ASIO data");

      FinalRes = E_NOINTERFACE;

      break;

    }

    HRESULT Res;

    AsioChannelParameters ChanPrm;

    Res = FilterCI->GetChannelParameters (ChanPrm);

    if (Res != S_OK) {

      Error (L"Cannot get ASIO channel parameters (%s)", ErrorText (Res));

      FinalRes = Res;

      break;

    }

    Assert (ChanPrm.NumChannels != 0 && ChanPrm.NumChannels <= NumAsioChannelsMax);

    SetCtlUInt (IDC_AsioChannels, ChanPrm.NumChannels);

    SetCtlText (IDC_AsioSampleType, SampleTypeList [ChanPrm.SampleFormats [0].FloatingPoint]);
    SetCtlUInt (IDC_AsioBps, ChanPrm.SampleFormats [0].BitsPerSample);
    SetCtlUInt (IDC_AsioBpc, ChanPrm.SampleFormats [0].BytesPerSample);
    SetCtlText (IDC_AsioSampleAlignment, SampleAlignmentList [ChanPrm.SampleFormats [0].RightAlign]);

    FinalRes = S_OK;

  } OnlyOnce;

  if (FinalRes != S_OK) {

    ClearControls (AsioChannelControls);

  }

  return FinalRes;

}





//
// Query ASIO filter and fill buffer info dialog controls
// =============================================================================================
//

HRESULT PopulateAsioBufferInfo () {

  HRESULT FinalRes;

  do {

    IDsOverAsio * FilterCI = (CaptureFilterCI != nullptr)? CaptureFilterCI : RenderFilterCI;

    if (FilterCI == nullptr) {

      Debug (6, L"No filter control interface to populate ASIO data");

      FinalRes = E_NOINTERFACE;

      break;

    }

    HRESULT Res;

    bool Translated;

    if (!Started) {

      UINT const SamplingRate = GetCtlUInt (IDC_SamplingRate, Translated);

      if (Translated && SamplingRate != 0) {

        Res = FilterCI->SetSamplingRate (SamplingRate);

      }

    }

    AsioBufferParameters BufPrm;

    Res = FilterCI->GetBufferParameters (BufPrm);

    if (Res != S_OK) {

      Error (L"Cannot get ASIO buffer parameters (%s)", ErrorText (Res));

      FinalRes = Res;

      break;

    }

    Assert (BufPrm.BufferSizes.Min != 0 && BufPrm.BufferSizes.Max != 0);
    Assert (BufPrm.BufferSizes.Preferred != 0);

    AssertIf (
      BufPrm.ActualBufferSize != 0,
      IsInRange (BufPrm.ActualBufferSize, BufPrm.BufferSizes.Min, BufPrm.BufferSizes.Max)
    );

    AssertIf (BufPrm.Latency != UINT (-1), BufPrm.Latency <= BufPrm.BufferSizes.Max);



    if (BufPrm.Latency != UINT (-1)) {

      SetCtlUInt (IDC_AsioLatency, BufPrm.Latency);

    } else {

      SetCtlText (IDC_AsioLatency, L"n/a");

    }

    SetCtlUInt (IDC_AsioBufferMin, BufPrm.BufferSizes.Min);
    SetCtlUInt (IDC_AsioBufferMax, BufPrm.BufferSizes.Max);
    SetCtlUInt (IDC_AsioBufferPref, BufPrm.BufferSizes.Preferred);
    SetCtlUInt (IDC_AsioBufferGran, BufPrm.BufferSizes.Granularity);
    SetCtlUInt (IDC_AsioBufferActual, BufPrm.ActualBufferSize);

    FinalRes = S_OK;

  } OnlyOnce;

  if (FinalRes != S_OK) {

    ClearControls (AsioBufferControls);

  }

  return FinalRes;

}





//
// Fill ASIO dialog controls
// =============================================================================================
//

HRESULT PopulateAsioParameters () {

  HRESULT FinalRes;

  do {

    HRESULT Res;

    Res = PopulateAsioChannelInfo ();

    if (Res != S_OK) {

      FinalRes = Res;

      break;

    }

    Res = PopulateAsioBufferInfo ();

    if (Res != S_OK) {

      FinalRes = Res;

      break;

    }

    FinalRes = S_OK;

  } OnlyOnce;

  return FinalRes;

}





//
// Select current ASIO capture device
// =============================================================================================
//

HRESULT SelectCurrentAsioCaptureDevice () {

  Assert (CaptureFilterCI != nullptr);

  HRESULT FinalRes;

  do {

    LRESULT LRes;

    LRes = DlgMsg (IDC_AsioDevice, CB_GETCURSEL);

    Assert (LRes != CB_ERR);

    UINT const Idx = static_cast <UINT> (LRes);

    Assert (Idx < NumAsioDevices);

    AsioDeviceEntry const & DE = AsioDevices [Idx];

    HRESULT Res;

    Res = CaptureFilterCI->SetAsioDriverWindow (MainDlg);

    if (Res != S_OK) {

      Error (L"Cannot set ASIO driver window %#x (%s)", MainDlg, ErrorText (Res));

      FinalRes = Res;

      break;

    }

    Debug (
      7,
      L"Asking capture filter to select ASIO device \"%s\" (%s)",
      DE.Name, GuidText (DE.ClassId)
    );

    Res = CaptureFilterCI->SelectAsioDevice (DE.ClassId);

    if (Res != S_OK) {

      Error (L"Cannot select ASIO device \"%s\" for capture filter (%s)", DE.Name, ErrorText (Res));

      FinalRes = Res;

      break;

    }

    Res = PopulateAsioParameters ();

    if (Res != S_OK) {

      FinalRes = Res;

      break;

    }

    FinalRes = S_OK;

  } OnlyOnce;

  return FinalRes;

}





//
// Select current ASIO render device
// =============================================================================================
//

HRESULT SelectCurrentAsioRenderDevice () {

  Assert (RenderFilterCI != nullptr);

  HRESULT FinalRes;

  do {

    LRESULT LRes;

    LRes = DlgMsg (IDC_AsioDevice, CB_GETCURSEL);

    Assert (LRes != CB_ERR);

    UINT const Idx = static_cast <UINT> (LRes);

    Assert (Idx < NumAsioDevices);

    AsioDeviceEntry const & DE = AsioDevices [Idx];

    HRESULT Res;

    Res = RenderFilterCI->SetAsioDriverWindow (MainDlg);

    if (Res != S_OK) {

      Error (L"Cannot set ASIO driver window %#x (%s)", MainDlg, ErrorText (Res));

      FinalRes = Res;

      break;

    }

    Debug (
      7,
      L"Asking the render filter to select ASIO device \"%s\" (%s)",
      DE.Name, GuidText (DE.ClassId)
    );

    Res = RenderFilterCI->SelectAsioDevice (DE.ClassId);

    if (Res != S_OK) {

      Error (L"Cannot select ASIO device \"%s\" for render filter (%s)", DE.Name, ErrorText (Res));

      FinalRes = Res;

      break;

    }

    Res = PopulateAsioParameters ();

    if (Res != S_OK) {

      FinalRes = Res;

      break;

    }

    FinalRes = S_OK;

  } OnlyOnce;

  return FinalRes;

}





HRESULT OpenAsioControlPanel () {

  IDsOverAsio * FilterCI = (CaptureFilterCI != nullptr)? CaptureFilterCI : RenderFilterCI;

  Assert (FilterCI != nullptr);

  HRESULT const Res = FilterCI->ShowAsioControlPanel ();

  if (Res == S_OK) {

    Debug (8, L"Returned from ASIO Control Panel");

  } else {

    Error (L"Cannot open ASIO Control Panel (%s)", ErrorText (Res));

  }

  return Res;

}





void CloseCaptureSide () {

  Assert (!Started && !PinsConnected);

  Debug (7, L"Closing capture side");

  ReleaseObject (CaptureFilter IfDebug (, L"capture filter %p", CaptureFilter));

  if (CaptureFilterCI != nullptr) {

    ReleaseObject (CaptureFilterCI IfDebug (, L"capture CI %p", CaptureFilterCI));
    CloseAsioDir ();

  }

  Debug (8, L"Capture side closed");

}





//
// Open the capture side
// =============================================================================================
//

HRESULT OpenCaptureSide () {

  Debug (7, L"Opening capture side");

  Assert (!Started && !PinsConnected);

  HRESULT FinalRes;

  do {

    HRESULT Res;

    Res = InstantiateCurrentCaptureFilter ();

    if (Res != S_OK) {

      FinalRes = Res;

      break;

    }

    if (CaptureFilterCI != nullptr) {
      
      if (NumAsioDevices == 0) {

        Res = FillAsioDevicesList ();

        if (Res != S_OK) {

          FinalRes = Res;

          break;

        }

      }

      Res = SelectCurrentAsioCaptureDevice ();

      if (Res != S_OK) {

        FinalRes = Res;

        break;

      }

    }

    Debug (8, L"Capture side open");

    FinalRes = S_OK;

  } OnlyOnce;

  if (FinalRes != S_OK) {

    CloseCaptureSide ();

  }

  return FinalRes;

}






void CloseRenderSide () {

  Assert (!Started && !PinsConnected);

  Debug (7, L"Closing render side");

  ReleaseObject (RenderFilter IfDebug (, L"render filter %p", RenderFilter));

  if (RenderFilterCI != nullptr) {

    ReleaseObject (RenderFilterCI IfDebug (, L"render CI %p", RenderFilterCI));
    CloseAsioDir ();

  }

  Debug (8, L"Render side closed");

}





//
// Open the render side
// =============================================================================================
//

HRESULT OpenRenderSide () {

  Debug (7, L"Opening render side");

  Assert (!Started && !PinsConnected);

  HRESULT FinalRes;

  do {

    HRESULT Res;

    Res = InstantiateCurrentRenderFilter ();

    if (Res != S_OK) {

      FinalRes = Res;

      break;

    }

    if (RenderFilterCI != nullptr) {
      
      if (NumAsioDevices == 0) {

        Res = FillAsioDevicesList ();

        if (Res != S_OK) {

          FinalRes = Res;

          break;

        }

      }

      Res = SelectCurrentAsioRenderDevice ();

      if (Res != S_OK) {

        FinalRes = Res;

        break;

      }

    }

    Debug (8, L"Render side open");

    FinalRes = S_OK;

  } OnlyOnce;

  if (FinalRes != S_OK) {

    CloseRenderSide ();

  }

  return FinalRes;

}





//
// Set filter parameters
// =============================================================================================
//

HRESULT SetFilterParameters (
  IDsOverAsio * const CI,
  UINT const AsioBS, UINT const SR, SampleFormatDesc const SF, UINT const SPB, UINT const BP
) {

  Debug (7, L"Setting parameters for filter control interface %p", CI);

  Assert (!Started);

  HRESULT FinalRes;

  do {

    HRESULT Res;

    Debug (7, L"Setting ASIO device buffer size %u", AsioBS);

    Res = CI->SetDeviceBufferSize (AsioBS);

    if (Res != S_OK) {

      Error (L"Cannot set ASIO device buffer size %u (%s)", AsioBS, ErrorText (Res));

      FinalRes = Res;

      break;

    }

    Debug (7, L"Setting filter sampling rate %u", SR);

    Res = CI->SetSamplingRate (SR);

    if (Res != S_OK) {

      Error (L"Cannot set filter sampling rate %u (%s)", SR, ErrorText (Res));

      FinalRes = Res;

      break;

    }

    Debug (7, L"Setting pin sample type %#x", SF);

    Res = CI->SetPinSampleType (SF);

    if (Res != S_OK) {

      Error (L"Cannot set pin sample type %#x (%s)", SF, ErrorText (Res));

      FinalRes = Res;

      break;

    }

    Debug (7, L"Setting samples per buffer %u and buffer parts %u", SPB, BP);

    Res = CI->SetPinBufferSize (SPB, BP);

    if (Res != S_OK) {

      Error (L"Cannot set %u samples per buffer and %u parts (%s)", SPB, BP, ErrorText (Res));

      FinalRes = Res;

      break;

    }

    FinalRes = S_OK;

  } OnlyOnce;

  return FinalRes;

}





//
// Stop the activity
// =============================================================================================
//

void Stop () {

  HRESULT Res;

  Debug (7, L"Stopping");

  if (MediaControl != nullptr) {

    FILTER_STATE State;

    Res = MediaControl->GetState (100, PointerCast (OAFilterState *, &State));

    if (Res == S_OK) {

      Debug (
        7,
        L"Graph state: %s",
        (State < NumArrayElems (StateNames))? StateNames [State] : L"invalid"
      );

    } else {

      Debug (9, L"Cannot get capture filter state (%s)", ErrorText (Res));

      State = State_Running;

    }

    if (State != State_Stopped) {

      Debug (8, L"Stopping the graph");

      // Can return S_FALSE if not all filters are ready to run

      Res = MediaControl->Stop ();

      if (Res != S_OK) {

        Error (L"Cannot stop the graph (%s)", ErrorText (Res));

      }

    }

  }



  Started = false;



  if (GraphBuilder != nullptr) {

    if (CaptureFilter != nullptr) {

#if 1

      FILTER_STATE State;

      Res = CaptureFilter->GetState (100, &State);

      if (Res == S_OK) {

        Debug (
          7,
          L"Capture filter state: %s",
          (State < NumArrayElems (StateNames))? StateNames [State] : L"invalid"
        );

      } else {

        Debug (9, L"Cannot get capture filter state (%s)", ErrorText (Res));

      }

      if (FAILED (Res) || State != State_Stopped) {

        Debug (7, L"Stopping capture filter %p", CaptureFilter);

        Res = CaptureFilter->Stop ();

        if (Res == S_OK) {

          Debug (8, L"Capture filter %p stopped", CaptureFilter);

        } else {

          Debug (9, L"Cannot stop capture filter %p (%s)", CaptureFilter, ErrorText (Res));

        }

      }

#endif

#if 1

      if (PinsConnected) {

        Assert (OutputPin != nullptr && InputPin != nullptr);

        Debug (7, L"Disconnecting output %p and input %p pins", OutputPin, InputPin);

        Res = OutputPin->Disconnect ();

        if (SUCCEEDED (Res)) {

          Debug (8, L"Output pin %p disconnected ", OutputPin);

        } else {

          Debug (
            9,
            L"Cannot disconnect output %p and input %p pins (%s)",
            OutputPin, InputPin, ErrorText (Res)
          );

        }

        PinsConnected = false;

      }

#endif

#if 1

      if (CaptureFilterAdded) {

        Debug (7, L"Removing capture filter %p from the graph", CaptureFilter);

        Res = GraphBuilder->RemoveFilter (CaptureFilter);

        if (Res == S_OK) {

          Debug (8, L"Capture filter %p removed from the graph", CaptureFilter);

        } else if (FAILED (Res)) {

          Debug (
            9,
            L"Cannot remove capture filter %p from the graph (%s)",
            CaptureFilter, ErrorText (Res)
          );

        }

        CaptureFilterAdded = false;

      }

#endif

    } // CaptureFilter != nullptr



#if 1

    if (RenderFilter != nullptr) {

      if (RenderFilterAdded) {

        Debug (7, L"Removing render filter %p from the graph", CaptureFilter);

        Res = GraphBuilder->RemoveFilter (RenderFilter);

        if (Res == S_OK) {

          Debug (8, L"Render filter %p removed from the graph", RenderFilter);

        } else if (FAILED (Res)) {

          Debug (
            9,
            L"Cannot remove render filter %p from the graph (%s)",
            RenderFilter, ErrorText (Res)
          );

        }

        RenderFilterAdded = false;

      }

    }

#endif

  } // GraphBuilder != nullptr

  ReleaseObject (OutputPin IfDebug (, L"output pin %p", OutputPin));
  ReleaseObject (InputPin IfDebug (, L"input pin %p", InputPin));

  ReleaseObject (MediaControl IfDebug (, L"IMediaControl %p", MediaControl));
  Verify (ReleaseObject (GraphBuilder IfDebug (, L"IGraphBuilder %p", GraphBuilder)) == 0);

  PopulateAsioParameters ();

  EnableNonRealtimeControls (true);
  SetCtlText (IDC_StartStop, L"Start");
  EnableControl (IDC_StartStop, true);

  Debug (8, L"Stopped");

}





//
// Start the activity
// =============================================================================================
//

HRESULT Start () {

  HRESULT FinalRes;

  do {

    HRESULT Res;

    Assert (!Started);

    EnableControl (IDC_StartStop, false);
    EnableNonRealtimeControls (false);



    if (CaptureFilterCI != nullptr || RenderFilterCI != nullptr) {

      bool Translated;

      UINT AsioSamplesPerBuffer = 0;

      wchar_t Text [100];

      GetCtlText (IDC_AsioSamplesPerBuffer, Text, NumArrayElems (Text));

      if (wcsicmp (Text, L"pref") != 0) {

        AsioSamplesPerBuffer = GetCtlUInt (IDC_AsioSamplesPerBuffer, Translated);

        if (!Translated || AsioSamplesPerBuffer == 0) {

          Error (L"Invalid number of samples per buffer for ASIO");

          FinalRes = E_INVALIDARG;

          break;

        }

      }

      UINT const SamplingRate = GetCtlUInt (IDC_SamplingRate, Translated);

      if (!Translated || SamplingRate == 0) {

        Error (L"Invalid sampling rate");

        FinalRes = E_INVALIDARG;

        break;

      }

      UINT const PinBps = GetCtlUInt (IDC_PinBitsPerSample, Translated);

      if (!Translated || PinBps == 0) {

        Error (L"Invalid number of bits per sample");

        FinalRes = E_INVALIDARG;

        break;

      }

      bool const PinFloat = (DlgMsg (IDC_PinSampleType, CB_GETCURSEL) == 1);

      Debug (7, L"Pin sample encoding: %s", SampleTypeList [PinFloat]);



      UINT const PinSamplesPerBuffer = GetCtlUInt (IDC_PinSamplesPerBuffer, Translated);

      if (!Translated || PinSamplesPerBuffer == 0) {

        Error (L"Invalid pin buffer size");

        FinalRes = E_INVALIDARG;

        break;

      }

      Debug (7, L"Samples per pin buffer: %u", PinSamplesPerBuffer);



      UINT const PinBufferParts = GetCtlUInt (IDC_PinBufferParts, Translated);

      if (!Translated || PinBufferParts == 0) {

        Error (L"Invalid number of pin buffer parts");

        FinalRes = E_INVALIDARG;

        break;

      }



      SampleFormatDesc const PinSampleType (PinBps, 0, PinFloat);



      if (CaptureFilterCI != nullptr) {

        Debug (7, L"Setting capture filter parameters");

        Res = SetFilterParameters (
          CaptureFilterCI,
          AsioSamplesPerBuffer, SamplingRate, PinSampleType, PinSamplesPerBuffer, PinBufferParts
        );

        if (Res != S_OK) {

          FinalRes = Res;

          break;

        }

      }

      if (RenderFilterCI != nullptr) {

        Debug (7, L"Setting render filter parameters");

        Res = SetFilterParameters (
          RenderFilterCI,
          AsioSamplesPerBuffer, SamplingRate, PinSampleType, PinSamplesPerBuffer, PinBufferParts
        );

        if (Res != S_OK) {

          FinalRes = Res;

          break;

        }

      }


    } // *FilterCI != nullptr



    Res = FindPcmPin (CaptureFilter, true, OutputPin);

    if (Res != S_OK) {

      FinalRes = Res;

      break;

    }



    Res = FindPcmPin (RenderFilter, false, InputPin);

    if (Res != S_OK) {

      FinalRes = Res;

      break;

    }



    Debug (7, L"Instantiating IGraphBuilder");

    Res = CreateObjectInstance (CLSID_FilterGraph, __uuidof (IGraphBuilder), GraphBuilder);

    if (Res != S_OK) {

		  Error (L"Cannot create graph builder (%s)", ErrorText (Res));

      FinalRes = Res;

      break;

	  }

    Debug (7, L"IGraphBuilder: %p", GraphBuilder);

    Assert (GraphBuilder != nullptr);



    Debug (7, L"Adding render filter to the graph");

    Res = GraphBuilder->AddFilter (RenderFilter, L"Render");

    if (Res != S_OK) {

      Error (L"Cannot add capture filter to the graph (%s)", ErrorText (Res));

      FinalRes = Res;

      break;

    }

    RenderFilterAdded = true;



    Debug (7, L"Adding capture filter to the graph");

    Res = GraphBuilder->AddFilter (CaptureFilter, L"Capture");

    if (Res != S_OK) {

      Error (L"Cannot add capture filter to the graph (%s)", ErrorText (Res));

      FinalRes = Res;

      break;

    }

    CaptureFilterAdded = true;



#if ConnectPins

    Debug (7, L"Connecting output pin %p to input pin %p", OutputPin, InputPin);

    Res = GraphBuilder->Connect (OutputPin, InputPin);

    if (Res != S_OK) {

      Error (L"Cannot connect pins (%s)", ErrorText (Res));

      FinalRes = Res;

      break;

    }

    PinsConnected = true;

#endif



#if UseMediaControl

    Debug (7, L"Instantiating IMediaControl");

    Res = GetObjectInterface (GraphBuilder, __uuidof (IMediaControl), MediaControl);

    if (Res != S_OK) {

		  Error (L"Cannot create media control (%s)", ErrorText (Res));

      FinalRes = Res;

      break;

	  }

    Debug (7, L"IMediaControl: %p", MediaControl);

    Assert (MediaControl != nullptr);

#endif



#if PauseBeforeRun

    Debug (8, L"Pausing the graph");

    Assert (MediaControl != nullptr);

    // Can return S_FALSE if not all filters have completed state transition

    Res = MediaControl->Pause ();

    if (FAILED (Res)) {

      Error (L"Cannot pause the graph (%s)", ErrorText (Res));

      FinalRes = Res;

      break;

    }

    if (Res == S_FALSE) {

      Debug (9, L"Not all filters have completed state transition");

    }

#endif



#if RunGraph

    Debug (8, L"Running the graph");

    Started = true;

    Assert (MediaControl != nullptr);

    // Can return S_FALSE if not all filters are ready to run

    Res = MediaControl->Run ();

    if (FAILED (Res)) {

      Error (L"Cannot run the graph (%s)", ErrorText (Res));

      FinalRes = Res;

      break;

    }

    if (Res == S_FALSE) {

      Debug (9, L"Not all filters in the graph are ready to run");

    }

#endif


    PopulateAsioParameters ();


    SetCtlText (IDC_StartStop, L"Stop");
    EnableControl (IDC_StartStop, true);

    Debug (8, L"Started");

    FinalRes = S_OK;

  } OnlyOnce;

  if (FAILED (FinalRes)) {

    Stop ();

  }

  return FinalRes;

}





//
// Main dialog function
// =============================================================================================
//

INT_PTR DialogProc (HWND const Dlg, UINT const Msg, WPARAM const wp, LPARAM const lp) {

  Unreferenced (lp);

  INT_PTR DlgRes = true;

  switch (Msg) {





    //
    // Initializing the dialog
    // =======================
    //

    case WM_INITDIALOG: {

      MainDlg = Dlg;             // Save window handle in global variable

      bool Success = false;

      do {

        HRESULT Res;

        EnableAsioControls (false);

        DlgMsg (IDC_AsioSamplesPerBuffer, CB_ADDSTRING, 0, reinterpret_cast <LPARAM> (L"Pref"));
        FillComboBox (IDC_AsioSamplesPerBuffer, SamplesPerBufferList);
        DlgMsg (IDC_AsioSamplesPerBuffer, CB_SETCURSEL, 0);

        FillComboBox (IDC_SamplingRate, SamplingRateList);
        SetCtlUInt (IDC_SamplingRate, SamplingRateDef);

        FillComboBox (IDC_PinSampleType, SampleTypeList);
        //DlgMsg (IDC_PinSampleType, CB_SELECTSTRING, 0);
        DlgMsg (
          IDC_PinSampleType, CB_SELECTSTRING,
          static_cast <WPARAM> (-1),
          reinterpret_cast <LPARAM> (SampleTypeList [PinSampleTypeDef.FloatingPoint])
        );

        FillComboBox (IDC_PinBitsPerSample, BitsPerSampleList);
        SetCtlUInt (IDC_PinBitsPerSample, PinSampleTypeDef.BitsPerSample);

        FillComboBox (IDC_PinSamplesPerBuffer, SamplesPerBufferList);
        SetCtlUInt (IDC_PinSamplesPerBuffer, PinSamplesPerBufferDef);

        FillComboBox (IDC_PinBufferParts, NumBufferPartsList);
        SetCtlUInt (IDC_PinBufferParts, PinBufferPartsDef);

        Res = EnumeratePcmFilters (true);

        if (Res != S_OK) {

          break;

        }

        FillFilterList (
          true,
__if_exists (CaptureFilterTpl) {
          CaptureFilterTpl
}
__if_not_exists (CaptureFilterTpl) {
          nullptr
}
        );

        Res = EnumeratePcmFilters (false);

        if (Res != S_OK) {

          break;

        }

        FillFilterList (
          false,
__if_exists (RenderFilterTpl) {
          RenderFilterTpl
}
__if_not_exists (RenderFilterTpl) {
          nullptr
}
        );
        
        OpenCaptureSide ();
        OpenRenderSide ();

        ShowWindow (Dlg, SW_SHOW);

        Debug (7, L"Main dialog initialized");

        Success = true;

      } OnlyOnce;

      if (!Success) {

        DlgRes = 9998;

        Debug (8, L"Aborting main dialog");

        Stop ();

        CloseCaptureSide ();
        CloseRenderSide ();

        EndDialog (MainDlg, 9999);

      }

      break;

    }





    //
    // Command messages
    // ================
    //

    case WM_COMMAND: {

      int const CtrlId = LOWORD (wp);
      int const CtrlNotif = HIWORD (wp);
      
      Unreferenced (CtrlNotif);

      Debug (5, L"Notification %X from ctrl %u", CtrlNotif, CtrlId);



      switch (CtrlId) {
        


        //
        // Close the dialog
        // ================
        //

        case IDCANCEL: {

          Stop ();

          CloseCaptureSide ();
          CloseRenderSide ();

          Debug (8, L"Closing main dialog");

          EndDialog (MainDlg, 1);

          break;

        }



        //
        // Capture filter selector
        // =======================
        //

        case IDC_CaptureFilter: {

          if (CtrlNotif == CBN_SELCHANGE) {

            Debug (8, L"Capture filter selection changed");

            CloseCaptureSide ();
            OpenCaptureSide ();

          }

          break;

        }



        //
        // Render filter selector
        // =======================
        //

        case IDC_RenderFilter: {

          if (CtrlNotif == CBN_SELCHANGE) {

            Debug (8, L"Render filter selection changed");

            CloseRenderSide ();
            OpenRenderSide ();

          }

          break;

        }



        //
        // ASIO device selector
        // ====================
        //

        case IDC_AsioDevice: {

          if (CtrlNotif == CBN_SELCHANGE) {

            Debug (8, L"ASIO device selection changed");

            if (CaptureFilterCI != nullptr) {

              SelectCurrentAsioCaptureDevice ();

            }

            if (RenderFilterCI != nullptr) {

              SelectCurrentAsioRenderDevice ();

            }

          }

          break;

        }



        //
        // ASIO Control Panel request
        // ==========================
        //

        case IDC_AsioControlPanel: {

          if (CtrlNotif == BN_CLICKED) {

            Debug (8, L"ASIO Control Panel requested");

            OpenAsioControlPanel ();

          }

          break;

        }



        //
        // Sampling rate selector
        // ======================
        //

        case IDC_SamplingRate: {

          if (CtrlNotif == EN_CHANGE || CtrlNotif == CBN_SELCHANGE || CtrlNotif == CBN_EDITCHANGE) {

            Debug (8, L"Sampling rate changed");

            PopulateAsioBufferInfo ();

          }

          break;

        }



        //
        // Sample type/format selector
        // ===========================
        //

        case IDC_PinSampleType: {

          if (CtrlNotif == CBN_SELCHANGE) {

            Debug (8, L"Pin sample type changed");

            if (DlgMsg (IDC_PinSampleType, CB_GETCURSEL) == 1) {

              SetCtlUInt (IDC_PinBitsPerSample, 32);

            }

          }

          break;

        }



        //
        // Start/stop the activity
        // =======================
        //

        case IDC_StartStop: {

          if (!Started) {

            Debug (8, L"Start command");

            HRESULT Res;

            do {

              Res = Start ();

              if (Res != S_OK) {

                break;

              }

              Assert (Started);

              Res = S_OK;

            } OnlyOnce;

            if (Res != S_OK) {

              Stop ();

              Assert (!Started);

            }

          } else {

            Debug (8, L"Stop command");

            Stop ();

            Assert (!Started);

          } // Start/stop

          DlgRes = false;

          break;

        }



        default: {

          DlgRes = false;

          break;

        }



      } // switch (CtrlId)


      break;

    } // WM_COMMAND



    default: {

      DlgRes = false;

      break;


    }



  } // switch (Msg)



  return DlgRes;

}





//
// Main process function
// =============================================================================================
//

int WINAPI WinMain (HINSTANCE Inst, HINSTANCE PrevInst, PSTR CmdLine, int ShowCmd) {

  Unreferenced (PrevInst && CmdLine && ShowCmd);

  IfDebug (SetDebugPrintLevel (AppDebugPrintLevel));

  AppInst = Inst;

  UINT AppRes = 999999;

#if _Debug

  ComDbg.SetPrintLevelOffset (2);

#endif

  do {

    HRESULT Res;



    Res = CoInitialize (nullptr);

    if (Res != S_OK) {

      Error (L"Cannot initialize COM (%s)", ErrorText (Res));

      AppRes = 9998;

      break;

    }


#if 0

    HRSRC const Rsc = FindResource (AppInst, MAKEINTRESOURCE (IDD_MainDialog), RT_DIALOG);

    if (Rsc == 0) {

      Res = HRESULT_FROM_WIN32 (GetLastError ());

      Debug (9, L"Cannot find dialog resource (%s)", ErrorText (Res));

      break;

    }

    Debug (6, L"Dialog resource handle: %#x", Rsc);

    DLGTEMPLATEEX const * const DT = PointerCast (DLGTEMPLATEEX const *, LoadResource (AppInst, Rsc));

    if (DT == nullptr) {

      Res = HRESULT_FROM_WIN32 (GetLastError ());

      Debug (9, L"Cannot load dialog template (%s)", ErrorText (Res));

      break;

    }

    Debug (6, L"Dialog template address: %p", DT);

    Debug (6, L"Dialog size: %u x %u DLUs", DT->cx, DT->cy);

    Verify (FreeResource (reinterpret_cast <HGLOBAL> (const_cast <DLGTEMPLATEEX *> ((DT)))) == 0);

#endif



    INT_PTR const DlgRes = DialogBox (
      AppInst, 
      MAKEINTRESOURCE (IDD_MainDialog),
      HWND_DESKTOP,
      DialogProc
    );

    if (DlgRes > 0) {

      AppRes = UINT (DlgRes - 1);

    } else {

      MessageBox (HWND_DESKTOP, L"Cannot create dialog", L"testapp", MB_ICONERROR | MB_OK);

      AppRes = 999;

    }


  } OnlyOnce;



  Assert (FilterDll == nullptr);
  Assert (DllGetClassObject_a == nullptr);
  Assert (DllSetDebugPrintLevel_a == nullptr);

  Assert (GraphBuilder == nullptr);
  Assert (MediaControl == nullptr);

  Assert (RenderFilter == nullptr && InputPin == nullptr);
  Assert (CaptureFilter == nullptr && OutputPin == nullptr);

  Assert (!PinsConnected);
  Assert (CaptureFilterCI == nullptr || RenderFilterCI == nullptr);

  Debug (6, L"Terminating COM");

  CoUninitialize ();

  Debug (7, L"COM terminated");

  Assert (AppRes != 999999);

  Debug (6, L"Exiting WinMain with %u", AppRes);

  return int (AppRes);

}
