#include "stdafx.h"





// #############################################################################################
//
// Variables
// =============================================================================================
//
// #############################################################################################





DebugPort ComDbg;





// #############################################################################################
//
// DebugPort class implementation
// =============================================================================================
//
// #############################################################################################





#if _Debug





void _cdecl DebugPort::DebugPrintf (UINT const Level, PCTSTR const _Fmt, ...) {

  UINT const ActualLevel = max (Level + PrintLevelOffset, 9);

  if (ActualLevel >= DebugPrintLevel) {

    va_list List;

    va_start (List, _Fmt);
    DebugVPrintf (Level, _Fmt, List);
    va_end (List);

  }

}





#endif _Debug





// #############################################################################################
//
// ErrorText_i class implementation
// =============================================================================================
//
// #############################################################################################





PCTSTR ErrorText_i::ToStr (PTSTR const Buf, UINT const MaxLen, DWORD const Code) {

  DWORD const LastErr = GetLastError ();

  PTSTR cp = Buf;
  UINT Avail = MaxLen;

  UINT Len;

  Len = static_cast <UINT> (
    _snwprintf (
      cp, Avail,
      TestBits (Code, 0xFFFF0000)? L"0x%08X - " : L"%u - ",
      Code
    )
  );

  cp += Len;
  Avail -= Len;

  Len = AMGetErrorText (static_cast <HRESULT> (Code), cp, Avail);

  if (Len == 0) {

    Len = FormatMessage (
      FORMAT_MESSAGE_FROM_SYSTEM | FORMAT_MESSAGE_IGNORE_INSERTS | FORMAT_MESSAGE_MAX_WIDTH_MASK,
      nullptr,
      Code,
      MAKELANGID (LANG_NEUTRAL, SUBLANG_DEFAULT),
      cp, DWORD (Avail),
      nullptr
    );

    if (Len == 0 && GetLastError () == ERROR_INSUFFICIENT_BUFFER) {

      __debugbreak ();

    }

  }

  if (Len != 0) {

    for (
      PTSTR p = cp + Len - 1;
      p > cp && (*p == L' ' || *p == L'\t' || *p == L'.' || *p == L'\n' || *p == L'\r');
      --p
    ) {

      *p = L'\0';

    }

  } else {

    Len = static_cast <UINT> (_snwprintf (cp, Avail, L"*** no error description ***"));

  }

  Assert (wcslen (Buf) < MaxLen);

  SetLastError (LastErr);

  return Buf;

}





// #############################################################################################
//
// GuidText_i class implementation
// =============================================================================================
//
// #############################################################################################





PTSTR GuidText_i::ToStr (PTSTR const Buffer, UINT const MaxLen, GUID const & Guid) {

  Assert (&Guid != nullptr);

  wchar_t Str [40];

  if (Guid != GUID_NULL) {

    wsprintf (
      Str,
      L"%08x-%04x-%04x-%02x%02x-%02x%02x%02x%02x%02x%02x",
      UINT (Guid.Data1),
      Guid.Data2, Guid.Data3,
      Guid.Data4 [0], Guid.Data4 [1],
      Guid.Data4 [2], Guid.Data4 [3], Guid.Data4 [4],
      Guid.Data4 [5], Guid.Data4 [6], Guid.Data4 [7]
    );

  } else {

    wcscpy (Str, L"GUID_NULL");

  }

  size_t const Len = wcslen (Str);
  size_t const ActualLen = min (Len, MaxLen - 1);

  memcpy (Buffer, Str, StrLenToSz (ActualLen));

  Buffer [ActualLen] = L'\0';

  return Buffer;

}





// #############################################################################################
//
// Global functions
// =============================================================================================
//
// #############################################################################################





#if _Debug





void SetDebugPrintLevel (UINT const Level) {

  DebugPrintLevel = Level;

}





void DebugVPrintf (UINT const Level, PCTSTR const _Fmt, va_list const List) {

  if (Level >= DebugPrintLevel) {

    wchar_t Buf [512];

    wchar_t const Delim = (Level == 9)? L'*' : L' ';

    int N = wsprintf (
      Buf,
      L"dstest %06u%c%4u%c",
      GetTickCount () % 100000, Delim, GetCurrentThreadId (), Delim
    );

    N += wvsprintf (Buf + N, _Fmt, List);

    wcscpy (Buf + N, L"\n");

    OutputDebugString (Buf);

  }

}





void _cdecl DebugPrintf (UINT const Level, PCTSTR const _Fmt, ...) {

  if (Level >= DebugPrintLevel) {

    va_list List;

    va_start (List, _Fmt);
    DebugVPrintf (Level, _Fmt, List);
    va_end (List);

  }

}





#endif // _DEBUG





//
// Common error message function
// =============================================================================================
//

void vError (PCTSTR const Format, va_list const Args) {

  DWORD const LastErr = GetLastError ();

  wchar_t Text [256];

  _vsnwprintf (Text, NumArrayElems (Text), Format, Args);

  Debug (9, L"Error: %.128s", Text);

  UINT const Flags = MB_ICONERROR | MB_OK;

  HWND const Win = (MainDlg != 0 && IsWindowVisible (MainDlg))? MainDlg : HWND_DESKTOP;

  MessageBox (Win, Text, L"Error", Flags);

  SetLastError (LastErr);

}





//
// Display error message without help capability
// =============================================================================================
//

void _cdecl Error (PCTSTR const Format, ...) {

  va_list Args;

  va_start (Args, Format);
  vError (Format, Args);
  va_end (Args);

}





//
// Get COM object reference counter
// =============================================================================================
//

UINT GetObjectRefCount (IUnknown * const Obj) {

  Assert (Obj != nullptr);

  Obj->AddRef ();

  return Obj->Release ();

}





//
// Check for template matching
// =============================================================================================
//
// '?' - any character
// '*' - any character sequence or empty
//

bool IsTemplateMatching (PCTSTR const Template, PCTSTR const Str, bool const IgnoreCase) {

  PCTSTR tp = Template;
  PCTSTR sp = Str;

  while (*tp != L'\0' || *sp != L'\0') {

    switch (*tp) {

      case L'?': {

        if (*sp != L'\0') {

          ++tp;
          ++sp;

          break;

        } else {

          return false;

        }

      }

      case L'*': {

        if (*++tp) {

          while (*sp != L'\0') {

            if (
              (IgnoreCase? (
                CharToUpper (*sp) == CharToUpper (*tp)
              ) : (
                *sp == *tp
              ))
              && IsTemplateMatching (tp, sp, IgnoreCase)
            ) {

              return true;

            }

            ++sp;

          }

          return false;

        } else {

          return true;

        }

      }

      default: {

        if (
          IgnoreCase? (
            CharToUpper (*tp++) != CharToUpper (*sp++)
          ) : (
            *tp++ != *sp++
          )
        ) {

          return false;

        }

      }

    }

  }

  return true;

}





LRESULT DlgMsg (UINT const CtrlId, UINT const Msg, WPARAM const WP, LPARAM const LP) {

  return SendDlgItemMessage (MainDlg, static_cast <int> (CtrlId), Msg, WP, LP);

}





UINT GetCtlText (UINT const Id, PTSTR const Text, UINT const MaxLen) {

  return GetDlgItemText (MainDlg, static_cast <int> (Id), Text, static_cast <int> (MaxLen));

}





bool SetCtlText (UINT const Id, PCTSTR const Text) {

  return !!SetDlgItemText (MainDlg, static_cast <int> (Id), Text);

}





bool SetCtlUInt (UINT const Id, UINT const Value) {

  return !!SetDlgItemInt (MainDlg, static_cast <int> (Id), Value, false);

}





UINT GetCtlUInt (UINT const Id, bool & Success) {

  BOOL Translated;

  UINT Val = GetDlgItemInt (MainDlg, static_cast <int> (Id), &Translated, false);

  if ((Success = !!Translated) != true) {

    Val = UINT (-1);

  }

  return Val;

}





void FillComboBox (UINT const Id, UINT const Values []) {

  for (UINT const * p = Values; *p != UINT_MAX; ++p) {

    wchar_t Text [16];
  
    wsprintf (Text, L"%u", *p);

    DlgMsg (Id, CB_ADDSTRING, 0, reinterpret_cast <LPARAM> (Text));

  }

}





void FillComboBox (UINT const Id, PCTSTR const Values []) {

  for (PCTSTR const * p = Values; *p != nullptr; ++p) {

    DlgMsg (Id, CB_ADDSTRING, 0, LPARAM (*p));

  }

}





void ClearControls (UINT const Ids []) {

  for (UINT const * p = Ids; *p != 0; ++p) {

    SetCtlText (*p, L"");

  }

}





bool EnableControl (UINT const Id, bool const Enable) {

  return !!EnableWindow (GetDlgItem (MainDlg, static_cast <int> (Id)), Enable);

}





void EnableControls (UINT const Ids [], bool const Enable) {

  for (UINT const * p = Ids; *p != 0; ++p) {

    EnableControl (*p, Enable);

  }

}





