// #############################################################################################
//
// Macros
// =============================================================================================
//
// #############################################################################################





// #############################################################################################
//
// Classes
// =============================================================================================
//
// #############################################################################################





struct DLGTEMPLATEEX {

  WORD dlgVer;
  WORD signature;
  DWORD helpID;
  DWORD exStyle;
  DWORD style;
  WORD cDlgItems;
  short x;
  short y;
  short cx;
  short cy;
  //sz_Or_Ord menu;
  //sz_Or_Ord windowClass;
  //WCHAR title[titleLen];
  //WORD pointsize;
  //WORD weight;
  //BYTE italic;
  //BYTE charset;
  //WCHAR typeface [stringLen];

};





class DebugPort {

  UINT PrintLevelOffset;

public:

#if _Debug

  void DebugPrintf (UINT const Level, PCTSTR const Fmt, ...);

#else

  void __forceinline DebugPrintf () { }

#endif

  void SetPrintLevelOffset (UINT const Offset) { PrintLevelOffset = Offset; }

};





template <size_t BufSize> class TextInPlace {

  public:

  // Buffer is defined public to allow derived classes to obtain its size via sizeof ().

  wchar_t Buf [BufSize];

  operator wchar_t const * () const { return Buf; }
  operator wchar_t * () { return Buf; }

};





class ErrorText_i : public TextInPlace <256> {

  public:

  ErrorText_i (DWORD const Code) {
    ToStr (Buf, NumArrayElems (Buf), Code);
  }

  // HRESULT is LONG too

  ErrorText_i (LONG const Code) {
    ToStr (Buf, NumArrayElems (Buf), DWORD (Code));
  }

  static PCTSTR ToStr (PTSTR const Buf, UINT const MaxLen, DWORD const Code);

};

// XxxText () functions must be macros to properly handle temporary objects

#define ErrorText(Code) (static_cast <PCTSTR> (ErrorText_i (Code)))





class GuidText_i : public TextInPlace <40> {

  public:

  GuidText_i (GUID const & Guid) {
    ToStr (Buf, NumArrayElems (Buf), Guid);
  }

  static PTSTR ToStr (PTSTR const Buffer, UINT const MaxLen, GUID const & Guid);

};

// XxxText () functions must be macros to properly handle temporary objects

#define GuidText(Guid) (static_cast <PCTSTR> (GuidText_i (Guid)))





// #############################################################################################
//
// Prototypes
// =============================================================================================
//
// #############################################################################################





void _cdecl Error (PCTSTR const Format, ...);

UINT GetObjectRefCount (IUnknown * const Obj);



void vError (PCTSTR const Format, va_list const Args);
void _cdecl Error (PCTSTR const Format, ...);
UINT GetObjectRefCount (IUnknown * const Obj);
bool IsTemplateMatching (PCTSTR const Template, PCTSTR const Str, bool const IgnoreCase = false);

LRESULT DlgMsg (UINT const CtrlId, UINT const Msg, WPARAM const WP = 0, LPARAM const LP = 0);

UINT GetCtlText (UINT const Id, PTSTR const Text, UINT const MaxLen);
UINT GetCtlUInt (UINT const Id, bool & Translated);
bool SetCtlText (UINT const Id, PCTSTR const Text);
bool SetCtlUInt (UINT const Id, UINT const Value);

void FillComboBox (UINT const Id, UINT const Values []);
void FillComboBox (UINT const Id, PCTSTR const Values []);
void ClearControls (UINT const Ids []);

bool EnableControl (UINT const Id, bool const Enable);
void EnableControls (UINT const Ids [], bool const Enable);





// #############################################################################################
//
// Functions
// =============================================================================================
//
// #############################################################################################





//
// Create a COM object instance
// =============================================================================================
//

template <typename T> HRESULT CreateObjectInstance (
  GUID const & ClsId, GUID const & IfId, T * & Obj
) {

  Assert (Obj == nullptr);

  HRESULT const Res = CoCreateInstance (
    ClsId, nullptr, CLSCTX_INPROC_SERVER, IfId,
    PointerCast (void **, &Obj)
  );

  if (Res == S_OK) {

    ComDbg.Debug (
      6,
      L"Created instance of %s/%s: %p",
      GuidText (ClsId), GuidText (IfId), Obj
    );

    Assert (Obj != nullptr);

  } else {

    ComDbg.Debug (
      9,
      L"Cannot create instance of %s/%s (%s)",
      GuidText (ClsId), GuidText (IfId), ErrorText (Res)
    );

    Assert (Obj == nullptr);

  }

  return Res;

}





//
// Get COM object interface
// =============================================================================================
//

template <typename T> HRESULT GetObjectInterface (
  IUnknown * const Src, GUID const IID, T * & Obj
) {

  Assert (Obj == nullptr);

  HRESULT const Res = Src->QueryInterface (IID, PointerCast (void **, &Obj));

  if (Res == S_OK) {

    Assert (Obj != nullptr);

    Verify (ReferenceObject (Obj) > 1);

    IfDebug (ULONG const R = ) Obj->Release ();

    ComDbg.Debug (6, L"Interface %s from %p: %p, %u refs", GuidText (IID), Src, Obj, R);

  } else {

    Assert (Obj == nullptr);

    ComDbg.Debug (9, L"No interface %s in %p", GuidText (IID), Obj);

  }

  return Res;

}





//
// Register a reference to a COM object
// =============================================================================================
//

template <typename T> UINT ReferenceObject (

  T * & Obj

#if _Debug

  , PCTSTR Fmt = nullptr, ...

#endif

) {

  Assert (&Obj != nullptr);

#if _Debug

  wchar_t Str [50];

  if (Fmt != nullptr) {

    va_list Args;

    va_start (Args, Fmt);

    _vsnwprintf (Str, NumArrayElems (Str), Fmt, Args);

    va_end (Args);

  } else {

    *Str = L'\0';

  }

  ComDbg.Debug (5, L"Referencing %s%s%p", Str, (*Str != L'\0')? L" " : L"", Obj);

#endif

  ULONG const R = Obj->AddRef ();

#if _Debug

  ComDbg.Debug (5, L"Referenced %s%%s%p, %u refs", Str, (*Str != L'\0')? L" " : L"", Obj, R);

#endif

  return R;

}





//
// Release a COM object and clear the pointer
// =============================================================================================
//

template <typename T> UINT ReleaseObject (

  T * & Obj

#if _Debug

  , PCTSTR Fmt = nullptr, ...

#endif
  
) {

  Assert (&Obj != nullptr);

  ULONG Remaining = 0;

  if (Obj != nullptr) {

#if _Debug

    wchar_t Str [50];

    if (Fmt != nullptr) {

      va_list Args;

      va_start (Args, Fmt);

      _vsnwprintf (Str, NumArrayElems (Str), Fmt, Args);

      va_end (Args);

    } else {

      *Str = L'\0';

    }

    ComDbg.Debug (5, L"Releasing %s%s%p", Str, (*Str != L'\0')? L" " : L"", Obj);

#endif

    Remaining = Obj->Release ();

#if _Debug

    ComDbg.Debug (
      5,
      L"Released %s%s%p, %u refs remaining",
      Str, (*Str != L'\0')? L" " : L"", Obj, Remaining
    );

#endif

    Obj = nullptr;

  }

  return Remaining;

}





extern DebugPort ComDbg;
